package com.tencent.subscribe.demo;

import com.tencent.subscribe.SubscribeDataProto;
import org.apache.commons.cli.*;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.*;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.header.Header;

import javax.xml.bind.DatatypeConverter;
import java.io.ByteArrayOutputStream;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.*;

public class TdsqlSubscribeFromKafka {
    private static Map<Integer, partitionMsgConsumer> partitionMsgConsumers = new HashMap<>();
    public static void main(final String[] args) {
        Option option_brokers = Option.builder()
                .required(true)
                .desc("The brokers option")
                .longOpt("brokers")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_topic = Option.builder()
                .required(true)
                .desc("The topic option")
                .longOpt("topic")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_group = Option.builder()
                .required(true)
                .desc("The group option")
                .longOpt("group")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_user = Option.builder()
                .required(false)
                .desc("The user option")
                .longOpt("user")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_password = Option.builder()
                .required(false)
                .desc("The password option")
                .longOpt("password")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_trans = Option.builder()
                .required(false)
                .desc("The trans2sql option")
                .longOpt("trans2sql")
                .build();

        Options options = new Options()
                .addOption(option_brokers)
                .addOption(option_topic)
                .addOption(option_user)
                .addOption(option_password)
                .addOption(option_group)
                .addOption(option_trans);

        CommandLineParser parser = new DefaultParser();

        try {
            CommandLine commandLine = parser.parse(options, args);

            run(
                    commandLine.getOptionValue("brokers"),
                    commandLine.getOptionValue("topic"),
                    commandLine.getOptionValue("group"),
                    commandLine.getOptionValue("user"),
                    commandLine.getOptionValue("password"),
                    commandLine.hasOption("trans2sql")
            );

        } catch (Exception exception) {
            System.out.print("error: ");
            System.out.println(exception.getMessage());
        }
    }

    private static void run(
            String brokers,
            String topic,
            String group,
            String user,
            String password,
            boolean trans2sql) throws Exception {
        Properties props = new Properties();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, brokers);
        props.put(ConsumerConfig.GROUP_ID_CONFIG, group);
        props.put(ConsumerConfig.CLIENT_ID_CONFIG, group + "_" + Instant.now().getEpochSecond());
        props.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        props.put(ConsumerConfig.ISOLATION_LEVEL_CONFIG, "read_committed");
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest");
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.ByteArrayDeserializer");

        if (null != user && !user.equals("")) {
            props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_PLAINTEXT");
            props.put(SaslConfigs.SASL_MECHANISM, "SCRAM-SHA-512");
            props.put(SaslConfigs.SASL_JAAS_CONFIG,
                    "org.apache.kafka.common.security.scram.ScramLoginModule required " +
                            "  username=\"" + user + "\"" +
                            "  password=\"" + password + "\";");
        }

        KafkaConsumer<String, byte[]> consumer = new KafkaConsumer<>(props);
        consumer.subscribe(Collections.singletonList(topic), new partitionSubLis());
        try {
            while (true) {
                ConsumerRecords<String, byte[]> records = consumer.poll(Duration.ofSeconds(1));
                for (ConsumerRecord<String, byte[]> record : records) {
                    partitionMsgConsumer pc = partitionMsgConsumers.get(record.partition());
                    if (null == pc) {
                        throw new IllegalStateException("no such partition consumer " + record.partition());
                    }

                    pc.consume(consumer, record, trans2sql);
                }
            }
        } finally {
            consumer.close();
        }
    }

    private final static String partitionSeqKey = "ps";

    private static long getPartitionSeq(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(partitionSeqKey).iterator();
        if (it.hasNext()) {
            Header h = it.next();
            return Long.parseUnsignedLong(new String(h.value()));
        }

        throw new IllegalStateException(partitionSeqKey + " does not exists");
    }


    private final static String ShardIdKey = "ShardId";

    private static String getShardId(ConsumerRecord<String, byte[]> msg) {
        Iterator<Header> it = msg.headers().headers(ShardIdKey).iterator();
        if (it.hasNext()) {
            Header h = it.next();
            return new String(h.value());
        }

        throw new IllegalStateException(ShardIdKey + " does not exists");
    }

    private static void onEntry(int partitionNum, long offset, long partitionSeq, SubscribeDataProto.Entry entry, boolean trans2sql) throws Exception {
        if (trans2sql) {
            trans2SQL(partitionNum, offset, partitionSeq, entry);
        } else {
            System.out.println("partition:" + partitionNum+", offset:"+ offset + ", partitionSeq:" + partitionSeq
                    + ", "+entry.toString());
        }
    }

    private static void trans2SQL(int partitionNum, long offset, long partitionSeq, SubscribeDataProto.Entry entry) throws Exception {
        SubscribeDataProto.Header header = entry.getHeader();
        System.out.println(
                String.format("-->[partition: %d, offset: %d, partitionSeq: %d] [%s:%d], happenedAt: %s",
                        partitionNum,
                        offset,
                        partitionSeq,
                        header.getFileName(), header.getPosition(),
                        LocalDateTime.ofInstant(Instant.ofEpochSecond(header.getTimestamp()),
                                TimeZone.getDefault().toZoneId())));

        switch (entry.getHeader().getMessageType()) {
            case HEARTBEAT:
                System.out.println("heartbeat: " + entry.getEvent().getHeartbeatEvent().getEpoch());
                break;
            case CHECKPOINT:
                SubscribeDataProto.CheckpointEvent evt = entry.getEvent().getCheckpointEvent();
                System.out.println("Checkpoint: gtid=" + evt.getSyncedGtid()
                        + ", file=" + evt.getFileName() + ", position=" + evt.getPosition());
                break;
            case BEGIN:
                System.out.println("BEGIN");
                break;
            case COMMIT:
                System.out.println("COMMIT");
                break;
            case ROLLBACK:
                System.out.println("ROLLBACK");
                break;
            case DDL:
                System.out.println(entry.getEvent().getDdlEvent().getSql());
                break;
            case DML:
                switch (entry.getEvent().getDmlEvent().getDmlEventType()) {
                    case INSERT:
                        transInsert(entry);
                        break;
                    case DELETE:
                        transDelete(entry);
                        break;
                    case UPDATE:
                        transUpdate(entry);
                        break;
                    default:
                        throw new IllegalStateException("unsupported dml event type");
                }
                break;
            default:
                throw new IllegalStateException("unsupported message type");
        }
    }

    private static void transInsert(SubscribeDataProto.Entry entry) throws Exception {
        SubscribeDataProto.DMLEvent dmlEvt = entry.getEvent().getDmlEvent();
        System.out.print(
                String.format(
                        "INSERT INTO `%s`.`%s` VALUES ",
                        entry.getHeader().getSchemaName(),
                        entry.getHeader().getTableName()));
        String rowSep = "";
        for (SubscribeDataProto.RowChange row : dmlEvt.getRowsList()) {
            System.out.print(rowSep);
            System.out.print('(');
            String sep = "";
            for (SubscribeDataProto.Data col : row.getNewColumnsList()) {
                System.out.print(sep);
                System.out.print(decode(col));
                sep = ", ";
            }
            System.out.print(')');
            rowSep = ", ";
        }

        System.out.println();
    }

    private static void transDelete(SubscribeDataProto.Entry entry) throws Exception {
        SubscribeDataProto.DMLEvent dmlEvt = entry.getEvent().getDmlEvent();
        String header = String.format(
                "DELETE FROM `%s`.`%s` WHERE ",
                entry.getHeader().getSchemaName(),
                entry.getHeader().getTableName());
        for (SubscribeDataProto.RowChange row : dmlEvt.getRowsList()) {
            System.out.print(header);
            String sep = "";
            for (int i = 0; i < row.getOldColumnsCount(); i++) {
                SubscribeDataProto.Data col = row.getOldColumns(i);
                if (col.getDataType() == SubscribeDataProto.DataType.NA) {
                    continue;
                }

                System.out.print(sep);
                SubscribeDataProto.Column colDef = dmlEvt.getColumns(i);
                System.out.print(String.format("`%s`", colDef.getName()));
                if (col.getDataType() == SubscribeDataProto.DataType.NIL) {
                    System.out.print(" IS NULL");
                } else {
                    System.out.print(String.format(" = %s", decode(col)));
                }
                sep = " AND ";
            }
            System.out.println(" LIMIT 1");
        }
    }

    private static void transUpdate(SubscribeDataProto.Entry entry) throws Exception {
        SubscribeDataProto.DMLEvent dmlEvt = entry.getEvent().getDmlEvent();
        String header = String.format(
                "UPDATE `%s`.`%s` SET ",
                entry.getHeader().getSchemaName(),
                entry.getHeader().getTableName());
        for (SubscribeDataProto.RowChange row : dmlEvt.getRowsList()) {
            System.out.print(header);
            String sep = "";
            for (int i = 0; i < row.getNewColumnsCount(); i++) {
                SubscribeDataProto.Data col = row.getNewColumns(i);
                if (col.getDataType() == SubscribeDataProto.DataType.NA) {
                    continue;
                }

                System.out.print(sep);
                SubscribeDataProto.Column colDef = dmlEvt.getColumns(i);
                System.out.print(String.format("`%s`", colDef.getName()));
                if (col.getDataType() == SubscribeDataProto.DataType.NIL) {
                    System.out.print(" IS NULL");
                } else {
                    System.out.print(String.format(" = %s", decode(col)));
                }
                sep = ", ";
            }

            System.out.print(" WHERE ");
            sep = "";
            for (int i = 0; i < row.getOldColumnsCount(); i++) {
                SubscribeDataProto.Data col = row.getOldColumns(i);
                if (col.getDataType() == SubscribeDataProto.DataType.NA) {
                    continue;
                }

                System.out.print(sep);
                SubscribeDataProto.Column colDef = dmlEvt.getColumns(i);
                System.out.print(String.format("`%s`", colDef.getName()));
                if (col.getDataType() == SubscribeDataProto.DataType.NIL) {
                    System.out.print(" IS NULL");
                } else {
                    System.out.print(String.format(" = %s", decode(col)));
                }
                sep = " AND ";
            }
            System.out.println(" LIMIT 1");
        }
    }

    private static String decode(SubscribeDataProto.Data data) throws Exception {
        switch (data.getDataType()) {
            case INT8:
            case INT16:
            case INT32:
            case INT64:
            case UINT8:
            case UINT16:
            case UINT32:
            case UINT64:
            case DECIMAL:
            case FLOAT32:
            case FLOAT64:
                return data.getSv();
            case STRING:
                return "_binary'" + new String(data.getBv().toByteArray()) + "'";
            case BYTES:
                return "x'" + DatatypeConverter.printHexBinary(data.getBv().toByteArray()) + "'";
            case NA:
                return "DEFAULT";
            case NIL:
                return "NULL";
            default:
                throw new IllegalStateException("unsupported data type");
        }
    }

    private static class partitionMsgConsumer {
        ByteArrayOutputStream completeMsg = new ByteArrayOutputStream();
        HashMap<String, ByteArrayOutputStream> shardMsgMap = new HashMap<>();
        HashMap<String, Long> shardLastSeqMap = new HashMap<>();
        TopicPartition tp;

        private partitionMsgConsumer(TopicPartition tp) {
            this.tp = tp;
        }

        private void consume(KafkaConsumer<String, byte[]> consumer, ConsumerRecord<String, byte[]> record, boolean trans2sql) throws Exception {
            // Each message has a sequence id in the partition, this id is continuous monotonic increasing,
            // these messages are duplicated if the id rotate;
            // There is unexpected message lost in kafka if the id is discontinuous,
            // have to reset the subscribe and client when it happened.
            // When the task is restarted, the sequence id will start from 1
            long ps = getPartitionSeq(record);

            // Each message has a shard id,
            // this id can be used to distinguish that the message belongs to which shard
            String shardId = getShardId(record);
            System.out.println("shardId:" + shardId);
            Long shardLastSeq = shardLastSeqMap.get(shardId);
            if (shardLastSeq == null || ps == 1) {
                shardLastSeqMap.put(shardId, (long) 1);
            } else if (ps <= shardLastSeq) {
                System.out.println("duplicated message: " + record);
                return;
            } else if (ps != shardLastSeq + 1 && shardLastSeq != 0) {
                throw new IllegalStateException("the partition seq discontinuous, last: " + shardLastSeq + ", current: " + ps);
            }

            shardLastSeqMap.put(shardId, ps);
            SubscribeDataProto.Envelope envelope = SubscribeDataProto.Envelope.parseFrom(record.value());
            if (1 != envelope.getVersion()) {
                throw new IllegalStateException(String.format("unsupported version: %d", envelope.getVersion()));
            }

            if (1 == envelope.getTotal()) {
                completeMsg = new ByteArrayOutputStream();
                envelope.getData().writeTo(completeMsg);
            } else {
                ByteArrayOutputStream shardMsg = shardMsgMap.get(shardId);
                if (null == shardMsg) {
                    shardMsg = new ByteArrayOutputStream();
                    envelope.getData().writeTo(shardMsg);
                    shardMsgMap.put(shardId, shardMsg);
                } else {
                    envelope.getData().writeTo(shardMsg);
                }
            }

            if (envelope.getIndex() < envelope.getTotal() - 1) {
                return;
            }

            SubscribeDataProto.Entries entries;
            if (1 == envelope.getTotal()) {
               entries = SubscribeDataProto.Entries.parseFrom(completeMsg.toByteArray());
            } else {
                entries = SubscribeDataProto.Entries.parseFrom(shardMsgMap.get(shardId).toByteArray());
                // Pay attention to memory recycling for shardMsgMap
                shardMsgMap.remove(shardId);
            }

            for (SubscribeDataProto.Entry entry : entries.getItemsList()) {
                onEntry(record.partition(), record.offset(), ps, entry, trans2sql);
            }
            // commit at checkpoint message
            if (entries.getItemsCount() > 0
                    && entries.getItems(0).getHeader().getMessageType() == SubscribeDataProto.MessageType.CHECKPOINT) {
                consumer.commitAsync(Collections.singletonMap(tp, new OffsetAndMetadata(record.offset() + 1)), null);
            }
        }
    }

    private static class partitionSubLis  implements ConsumerRebalanceListener {
        public void onPartitionsAssigned(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.computeIfAbsent(p.partition(), k -> new partitionMsgConsumer(p));
            }
        }

        public void onPartitionsRevoked(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.remove(p.partition());
            }
        }
    }
}
